include Okauth_client_intf.Type

module In_memory_storage = struct
  type t = (string, string) Hashtbl.t

  let empty =
    Hashtbl.create 0

  let size =
    Hashtbl.length

  let find storage ~key =
    Hashtbl.find_opt storage key

  let set storage ~key ~value =
    Hashtbl.add storage key value

  let remove storage ~key =
    Hashtbl.remove storage key
end

module Make(C: module type of Okweb.Client)(S: Storage) = struct
  open Core
  open Okweb
  open Lwt.Infix
  open Sexplib.Std
  open Okauth_crypto

  type session = {
    s_user_id: string;
    s_expires: Core.Time.t;
  } [@@deriving sexp]

  type confirmation = {
    c_user_id: string;
    c_expires: Core.Time.t;
  } [@@deriving sexp]

  let default_session_duration =
    Core.Time.Span.of_day 14.0

  let default_confirmation_duration =
    Core.Time.Span.of_min 10.0

  let session_cookie_key =
    "session_id"

  let user_cookie_key =
    "user_id"

  let create_unique config value =
    let max_retries = 10 in
    let gen_key () = Uuid.(create_random Random.State.default |> to_string) in
    let key_opt =
      List.range 0 max_retries
      |> List.fold_until ~init:(gen_key ()) ~finish:(Fn.const None)
        ~f:begin fun key _ ->
          match S.find config.storage ~key with
          | None -> Continue_or_stop.Stop (Some key)
          | Some _ -> Continue_or_stop.Continue (gen_key ())
        end in
    match key_opt with
    | Some key ->
      S.set config.storage ~key ~value;
      key
    | None ->
      failwith "unable to generate unique key"

  let create_session config user_id =
    let session = {
      s_user_id = user_id;
      s_expires = Core.Time.(add (now ()) default_session_duration)
    } in
    create_unique config (sexp_of_session session |> Core.Sexp.to_string)

  let create_confirmation config user_id =
    let confirmation = {
      c_user_id = user_id;
      c_expires = Core.Time.(add (now ()) default_confirmation_duration)
    } in
    create_unique config (sexp_of_confirmation confirmation |> Core.Sexp.to_string)

  let generate_confirmation_url config uid host user_pek =
    let full_user_id = uid ^ "@" ^ host in
    let confirmation_key = create_confirmation config full_user_id in
    let temp_url = Printf.sprintf "%s/login/confirm/%s" config.base_url confirmation_key in
    let nonce = generate_nonce () in
    let payload = encrypt (Plain_text.of_string temp_url) user_pek config.secret_encryption_key nonce in
    let signature = sign (Cipher_text.to_string payload ^ Nonce.to_string nonce) config.secret_signing_key in
    Printf.sprintf
      "https://%s/v1/%s/authenticate?k=%s&n=%s&p=%s&s=%s"
      host
      uid
      (Public_key.to_base64 config.public_encryption_key)
      (Nonce.to_base64 nonce)
      (Cipher_text.to_base64 payload)
      (Signature.to_base64 signature)

  let do_login config full_user_id =
    match String.split ~on:'@' full_user_id with
    | [ uid; host ] ->
      (Printf.sprintf "https://%s/v1/%s/public_key" host uid)
      |> Request.get
      |> C.send
      >|= Response.get_body
      >>= Body.read_up_to
      >|= Public_key.of_base64_exn
      >|= generate_confirmation_url config uid host
      >|= Response.redirect
    | _ ->
      Lwt.return Response.(of_status 400 |> with_string_body "invalid user_id")

  let do_logout config req =
    let make_removal_cookie key =
      Cookie.make (key, "")
        ~expires:Core.Time.(of_span_since_epoch (Span.of_sec 0.0)) in
    Request.get_cookies req
    |> Cookie_jar.find_value ~key:session_cookie_key
    |> Option.iter ~f:(fun key -> S.remove config.storage ~key);
    Lwt.return Response.(
        redirect (config.base_url ^ "/login")
        |> with_cookies (Cookie_jar.of_list [
            make_removal_cookie session_cookie_key;
            make_removal_cookie user_cookie_key;
          ]))

  let has_valid_auth_cookie config req =
    let cookies = Request.get_cookies req in
    let cookie_user_id = Cookie_jar.find_value cookies ~key:user_cookie_key |> Option.value ~default:"" in
    Cookie_jar.find_value cookies ~key:session_cookie_key
    |> Option.bind ~f:(fun key -> S.find config.storage ~key)
    |> Option.map ~f:(fun v -> Core.Sexp.of_string v |> session_of_sexp)
    |> Option.filter ~f:(fun s -> Core.Time.is_later ~than:(Core.Time.now ()) s.s_expires)
    |> Option.filter ~f:(fun s -> String.equal s.s_user_id cookie_user_id)
    |> Option.is_some

  let confirm_login config confirmation_key =
    let confirmation_opt =
      S.find config.storage ~key:confirmation_key
      |> Option.map ~f:(fun v -> Core.Sexp.of_string v |> confirmation_of_sexp)
      |> Option.filter ~f:(fun c -> Core.Time.is_later ~than:(Core.Time.now ()) c.c_expires) in
    match confirmation_opt with
    | Some { c_user_id; _} ->
      S.remove config.storage ~key:confirmation_key;
      let session_key = create_session config c_user_id in
      let cookies = Cookie_jar.of_list [
          Cookie.make
            ~expires:Core.Time.(add (now ()) default_session_duration)
            ~path:"/"
            ~secure:true
            ~http_only:true
            (session_cookie_key, session_key);
          Cookie.make
            ~path:"/"
            ~expires:Core.Time.(add (now()) (Span.of_day 365.0))
            (user_cookie_key, c_user_id);
        ] in
      Lwt.return Response.(redirect config.base_url |> with_cookies cookies)
    | None ->
      Lwt.return Response.(of_status 400 |> with_string_body "confirmation failed")

  let set_cache_forever_header resp =
    resp |> Response.with_header ~key:"Cache-Control" ~value:"public, immutable, max-age=31536000"

  let set_content_type typ resp =
    resp |> Response.with_header ~key:"Content-Type" ~value:typ

  let get_login_page config =
    let template = Mustache.of_string [%blob "web/login_page.html"] in
    let params = `O [ "app_name", `String config.app_name ] in
    Response.(
      ok
      |> set_content_type "text/html; charset=utf-8"
      |> set_cache_forever_header
      |> with_string_body (Mustache.render template params))

  let get_okauth_info config =
    let info = Printf.sprintf "\
        Name: %s\n\
        Encryption Key: %s\n\
        Signing Key: %s"
        config.app_name
        (Public_key.to_base64 config.public_encryption_key)
        (Public_key.to_base64 config.public_signing_key) in
    Response.(
      ok
      |> set_content_type "text/plain"
      |> set_cache_forever_header
      |> with_string_body info)

  let do_middleware config handler req =
    match (Request.get_http_method req, Request.get_path req) with
    | ("GET", [ ".well-known"; "okauth.txt" ]) ->
      Lwt.return (get_okauth_info config)
    | ("GET", ["login"]) ->
      begin match req |> Request.get_cookies |> Cookie_jar.find_value ~key:user_cookie_key with
        | Some user_id -> do_login config user_id
        | None -> Lwt.return (get_login_page config)
      end
    | ("POST", ["login"]) ->
      req |> Request.get_body |> Form.of_body >|= Form.find_param ~key:"user_id" >>= begin function
        | Some user_id -> do_login config user_id
        | None -> Lwt.return Response.(of_status 400 |> with_string_body "invalid request")
      end
    | ("GET", ["login"; "confirm"; confirmation]) ->
      confirm_login config confirmation
    | ("GET", ["logout"]) ->
      do_logout config req
    | _ when has_valid_auth_cookie config req ->
      handler req
    | _ ->
      Lwt.return (Response.redirect (Printf.sprintf "%s/login" config.base_url))

  let middleware config handler =
    Looksok.middleware (do_middleware config handler)
end
